# -*- coding: utf-8; indent-tabs-mode: nil -*-

# mailbridge: a mail storage spanning bridge
# 
# Copyright 2011 Ryan Hughes <ryan@n-h.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging
import traceback
import threading
import threadpool
import select
import sys


class MailLogMonitor2DelAdviseSink(object):
    def __init__(self, updater, collection):
        ECChangeAdviseSink.__init__(self, [IID_IECChangeAdviseSink])
        self.updater = updater
        self.collection = collection
        self.lock = threading.Lock()
        self.synclock = threading.Lock()
        self.monitored = False
        self.events = {}

    def OnNotify(self, flags, entries):
        if len(entries) > 0:
            states = (SyncState(entry) for entry in entries)
            newevents = dict([state.syncid, state.changeid] for state in states)
            self.lock.acquire()
            exc = True
            try:
                try:
                    if self.monitored == True:
                        self.collection.AddEvents(self, newevents)
                    else:
                        self.events.update(newevents)
                except:
                    exc = False
                    if not self.lock.release():
                        raise
            finally:
                if exc:
                    self.lock.release()
        return 0
    
    def EnableMonitor(self):
        self.lock.acquire()
        exc = True
        try:
            try:
                if (self.monitored == False):
                    self.monitored = True
                    self.collection.AddEvents(self, self.events)
                    self.events = None
            except:
                exc = False
                if not self.lock.release():
                    raise
        finally:
            if exc:
                self.lock.release()

    def UpdateSyncState(self, state):
        self.collection.UpdateState(self, state)

    def SynchronizeFolder(self, syncid):
        self.synclock.acquire()
        exc = True
        try:
            try:
                self.updater.update_imap_cache_for_folder_by_syncid(syncid)
            except:
                exc = False
                if not self.synclock.release():
                    raise
        finally:
            if exc:
                self.synclock.release()


class MaillogMonitor:
    def __init__(self, maillog_pipe_path):
        self.condition = threading.Condition()
        self.doExit = False
        self.sinkevents = {}
        self.logger = logging.getLogger("MaillogMonitor")
        self.pipein = open(maillog_pipe_path, 'r')
        self.updater = {}

    def GetChangeAdviseSink(self, updater):
        return AdviseSink(updater, self)
    
    def addUpdater(self, account, updater):
        self.updater[account] = updater

    def AddEvents(self, sink, newevents):
        self.logger.debug("Adding %u new events from sink %s" % (len(newevents), sink.__hash__))
        self.condition.acquire()
        exc = True
        try:
            try:
                if (self.sinkevents.has_key(sink)):
                    events = self.sinkevents[sink]
                else:
                    self.sinkevents[sink] = events = {}
                events.update(newevents)
                self.condition.notify()
            except:
                exc = False
                if not self.condition.release():
                    raise
        finally:
            if exc:
                self.condition.release()
    
    def UpdateState(self, sink, state):
        self.logger.debug("Updating state for sink %s, syncid=%u, changeid=%u" % (sink.__hash__, state.syncid, state.changeid))
        self.condition.acquire()
        exc = True
        try:
            try:
                events = self.sinkevents.get(sink, None)
                if events and events.has_key(state.syncid) and events[state.syncid] <= state.changeid:
                    del events[state.syncid]
                    if not events:
                        del self.sinkevents[sink]
            except:
                exc = False
                if not self.condition.release():
                    raise
        finally:
            if exc:
                self.condition.release()
    
    def Cancel(self):
        self.logger.debug("Canceling MaillogMonitor...")
        self.condition.acquire()
        exc = True
        try:
            try:
                self.doExit = True
                self.condition.notify()
            except:
                exc = False
                if not self.condition.release():
                    raise
        finally:
            if exc:
                self.condition.release()
    
    def getInfoLine(self, infoline):
        if (infoline.find(':') > -1):
            action, paramsstr = infoline.split(':', 1)
            params = paramsstr.split(',')
            paramsmap = None
            infovalues = None
            if (len(params) > 1):
                paramsmap = {}
                for param in params:
                    key, value = param.split('=', 1)
                    paramsmap[key.strip()] = value
            else:
                infostr = params[0]
                if (infostr.find(' -> ') > -1):
                    infovalues = infostr.split(' -> ')
            return (action.strip(), paramsmap, infovalues)
        
    def Run(self, pool):
        self.logger.debug("MaillogMonitor started")
        exc = True
        inputs = [self.pipein]
        outputs = []
        timeout = 5
        try:
            try:
                while not self.doExit:
                    try:
                        readable, writable, exceptional = select.select(inputs, outputs, inputs, timeout)
                        if not (readable or writable or exceptional):
                            self.logger.debug("Timeout MailLog")
                            continue
                        lines = []
                        while (not self.doExit and (readable or writable or exceptional) and (readable[0] == self.pipein)):
                            line = self.pipein.readline()
                            lines.append(line)
                            readable, writable, exceptional = select.select(inputs, outputs, inputs, 0.2)
        
                        if not self.doExit:
                            self.logger.debug("Collected %d events for MailLog" % len(lines))
                            seen_boxes = {}
                            for line in lines:
                                self.logger.debug("Get events for MailLog: %s" % line)
                                month, day, time, hostname, rest = line.split(' ', 4)
                                process, service, rest = rest.split(':', 2)
                                service = service.strip()
                                if (service[:5] == 'imap('):
                                    user = service[5:-1]
                                    if not self.updater.has_key(user):
                                        continue
                                    infos = self.getInfoLine(rest)
                                    if (infos != None):
                                        action, paramsmap, infovalues = infos
                                        if (action == 'Mailbox renamed'):
                                            box = ''
                                        elif (action == 'Mailbox created'):
                                            box = ''
                                        else:
                                            if (paramsmap != None and paramsmap.has_key('box')):                                            
                                                box = paramsmap['box']
                                            else:
                                                continue
                                        if (not seen_boxes.has_key(user + ':' + box)):
                                            pool.putRequest(threadpool.WorkRequest(self.ProcessSink, [(user, box)], None))
                                            self.logger.debug("put Request for %s event for MailLog: %s" % (user + ':' + box, line))
                                            seen_boxes[user + ':' + box] = 1
                    except:
                        traceback.print_exc(file=sys.stdout)
            except:
                exc = False
        finally:
            pass
    
    def ProcessSink(self, args):
        user, box = args
        self.logger.debug("Processing events for box %s" % (box))
        if (box == ''):
            self.updater[user].check_update_backend_hierarchy()
        else:        
            self.updater[user].update_backend_for_folder_by_folder_path(box)
        self.logger.debug("Processed  events for box %s" % (box))
        